package de.ozgcloud.xta.test.app.data;

import java.math.BigInteger;
import java.time.ZonedDateTime;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;
import org.springframework.stereotype.Component;

import de.ozgcloud.xta.test.app.model.XtaMessage;
import de.ozgcloud.xta.test.app.model.XtaTransportReport;

@Component
@Log4j2
@RequiredArgsConstructor
public class XtaMessageRepository {

	private final Map<String, XtaMessage> messages = new HashMap<>();
	private final HashMap<String, XtaTransportReport.MessageStatus> messageIdToStatus = new HashMap<>();

	public boolean add(XtaMessage message) {
		if (messages.containsKey(message.metaData().messageId())) {
			return false;
		}
		messages.put(message.metaData().messageId(), message);
		messageIdToStatus.put(message.metaData().messageId(), XtaTransportReport.MessageStatus.OFFEN);
		return true;
	}

	public Optional<XtaMessage> getMessage(String messageId) {
		return Optional.ofNullable(messages.get(messageId));
	}

	public Optional<XtaTransportReport> getTransportReport(String messageId) {
		return Optional.ofNullable(messages.get(messageId))
				.map(m -> XtaTransportReport.builder()
						.messageMetaData(messages.get(messageId).metaData())
						.status(messageIdToStatus.get(messageId))
						.reportTime(getNow())
						.build());
	}

	public List<XtaMessage> findOpenMessagesByReaderIdentifierValueLimitedTo(String readerIdentifierValue, int maxListItems) {
		return findOpenMessagesByReaderIdentifierValue(readerIdentifierValue)
				.limit(maxListItems)
				.toList();
	}

	public BigInteger countOpenMessagesByReaderIdentifierValue(String readerIdentifierValue) {
		return BigInteger.valueOf(findOpenMessagesByReaderIdentifierValue(readerIdentifierValue).count());
	}

	Stream<XtaMessage> findOpenMessagesByReaderIdentifierValue(String readerIdentifierValue) {
		return messages.values().stream()
				.filter(m -> messageIdToStatus.get(m.metaData().messageId()) == XtaTransportReport.MessageStatus.OFFEN)
				.filter(m -> m.metaData().readerIdentifier().value().equals(readerIdentifierValue));
	}

	public boolean closeByMessageIdAndReaderIdentifierValue(String messageId, String readerIdentifierValue) {
		return getMessage(messageId)
				.filter(m -> m.metaData().readerIdentifier().value().equals(readerIdentifierValue))
				.map(m -> close(messageId))
				.orElse(false);
	}

	boolean close(String messageId) {
		return messageIdToStatus.computeIfPresent(messageId, (k, v) -> XtaTransportReport.MessageStatus.GRUEN) != null;
	}

	private ZonedDateTime getNow() {
		return ZonedDateTime.now();
	}
}
