#!/bin/bash

set -e

STORE_PASS=password
ISSUER_ALIAS=xta-test-root-ca
TRUST_STORE_JKS=xta-test_truststore.jks

if [ ! -f $ISSUER_ALIAS.key ] || [ ! -f $ISSUER_ALIAS.crt ]; then
  rm $TRUST_STORE_JKS || true
  echo "[1.0] Generate key for the Xta-Root-CA (Root CA key or crt not found)"
  openssl genrsa -out $ISSUER_ALIAS.key 2048
  echo "[1.1] Generate a self-signed certificate for the Xta-Root-CA"
  openssl req -x509 -new -nodes -key $ISSUER_ALIAS.key -sha256 -days 4000 -out $ISSUER_ALIAS.crt -config ca-openssl.cnf -extensions v3_ca

  # echo "[1.2] Generate a keystore for the Xta-Root-CA"
  # openssl pkcs12 -export -in $ISSUER_ALIAS.crt -inkey $ISSUER_ALIAS.key -out $ISSUER_ALIAS.p12 -name xta-test-server -passout pass:password
else
  echo "[1.0] Root CA found. Skipping generation."
fi

if [ ! -f $TRUST_STORE_JKS ]; then
  echo "[2.0] Import Root CA into Xta-Server-Truststore"
  keytool -importcert -alias $ISSUER_ALIAS -keystore $TRUST_STORE_JKS -storetype JKS -storepass "$STORE_PASS" -file $ISSUER_ALIAS.crt -noprompt
else
  echo "[2.0] Xta-Server-Truststore found. Skipping generation."
fi

function generate_keystore_with_signed_certificate {
  local step_num="$1"
  local key_alias="$2"
  local key_cn="$3"
  local ext_config_file="$4"
  local keystore_args=( -keystore "${key_alias}_keystore.p12" -storepass "$STORE_PASS" -storetype PKCS12 )
  rm "$key_alias"_keystore.p12 || true
  echo "[$step_num.0] Generate a keystore for $key_cn"
  keytool -genkeypair "${keystore_args[@]}" -alias "$key_alias" -keyalg RSA -keysize 2048 -validity 3900 -dname "CN=$key_cn, OU=XtaTestOrgUnit, O=XtaTestOrg, L=XtaTestCity, S=XtaTestState, C=DE"
  keytool -importcert "${keystore_args[@]}" -alias $ISSUER_ALIAS -file $ISSUER_ALIAS.crt -noprompt

  echo "[$step_num.1] Generate a certificate signing request for $key_cn"
  keytool -certreq "${keystore_args[@]}" -alias "$key_alias" -file "$key_alias.csr"

  echo "[$step_num.2] Sign the certificate with the Root CA using $ext_config_file"
  openssl x509 -req -in "$key_alias.csr" -out "$key_alias.crt" -CA $ISSUER_ALIAS.crt -CAkey $ISSUER_ALIAS.key -CAcreateserial -days 3900 -sha256 -extfile "$ext_config_file" -extensions req_ext
  rm "$key_alias.csr" || true

  echo "[$step_num.3] Import the signed certificate into the keystore with alias $key_alias"
  keytool -importcert "${keystore_args[@]}" -alias "$key_alias" -file "$key_alias.crt" -noprompt
  rm "$key_alias.crt" || true
  keytool -delete "${keystore_args[@]}" -alias $ISSUER_ALIAS -noprompt
}

generate_keystore_with_signed_certificate "3" xta-test-server "XTA Test Server" server-openssl.cnf
generate_keystore_with_signed_certificate "4" xta-test-client-john-smith "XTA Test Client John Smith" client-openssl.cnf
generate_keystore_with_signed_certificate "5" xta-test-client-jane-doe "XTA Test Client Jane Doe" client-openssl.cnf