package de.ozgcloud.xta.test.app.data;

import static de.ozgcloud.xta.test.app.factory.XtaMessageMetaDataTestFactory.READER_IDENTIFIER;
import static de.ozgcloud.xta.test.app.factory.XtaMessageTestFactory.MESSAGE_ID_1;
import static de.ozgcloud.xta.test.app.factory.XtaMessageTestFactory.MESSAGE_ID_2;
import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.Mockito.doReturn;

import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Spy;

import de.ozgcloud.xta.test.app.factory.XtaMessageMetaDataTestFactory;
import de.ozgcloud.xta.test.app.factory.XtaMessageTestFactory;
import de.ozgcloud.xta.test.app.model.XtaIdentifier;
import de.ozgcloud.xta.test.app.model.XtaMessage;
import de.ozgcloud.xta.test.app.model.XtaTransportReport;

public class XtaMessageRepositoryTest {

	@Spy
	@InjectMocks
	private XtaMessageRepository xtaMessageRepository;
	private final XtaMessage message = XtaMessageTestFactory.create(MESSAGE_ID_1);

	@DisplayName("save")
	@Nested
	class TestSave {

		@Test
		void shouldSaveMessage() {
			xtaMessageRepository.add(message);

			assertThat(xtaMessageRepository.getMessage(MESSAGE_ID_1)).contains(message);
		}
	}

	@DisplayName("get message")
	@Nested
	class TestGetMessage {
		@BeforeEach
		void beforeEach() {
			xtaMessageRepository = new XtaMessageRepository();
			xtaMessageRepository.add(message);
		}

		@Test
		void shouldReturnMessage() {
			var result = xtaMessageRepository.getMessage(MESSAGE_ID_1);

			assertThat(result).contains(message);
		}

		@Test
		void shouldReturnEmpty() {
			var result = xtaMessageRepository.getMessage(MESSAGE_ID_2);

			assertThat(result).isEmpty();
		}
	}

	@DisplayName("get transport report")
	@Nested
	class TestGetTransportReport {
		@Nested
		class WithExistingMessageId {
			@BeforeEach
			void beforeEach() {
				xtaMessageRepository = new XtaMessageRepository();
				xtaMessageRepository.add(message);
			}

			@Test
			void shouldReturnReport() {
				var result = xtaMessageRepository.getTransportReport(message.metaData().messageId());

				assertThat(result).isPresent();
				assertThat(result.get().messageMetaData()).isEqualTo(message.metaData());
			}

			@Test
			void shouldSetReportTime() {
				var result = xtaMessageRepository.getTransportReport(message.metaData().messageId());

				assertThat(result).isPresent();
				assertThat(result.get().reportTime()).isNotNull();
			}

			@Test
			void shouldSetStatusOffen() {
				var result = xtaMessageRepository.getTransportReport(message.metaData().messageId());

				assertThat(result).isPresent();
				assertThat(result.get().status()).isEqualTo(XtaTransportReport.MessageStatus.OFFEN);
			}
		}

		@Nested
		class WithMissingMessageId {
			@Test
			void shouldReturnEmpty() {
				var result = xtaMessageRepository.getTransportReport(message.metaData().messageId());

				assertThat(result).isEmpty();
			}
		}

	}

	@DisplayName("close by message id and reader identifier value")
	@Nested
	class TestCloseByMessageIdAndReaderIdentifierValue {
		@BeforeEach
		void beforeEach() {
			xtaMessageRepository = new XtaMessageRepository();
			xtaMessageRepository.add(message);
		}

		@Test
		void shouldCloseMessage() {
			var result = xtaMessageRepository.closeByMessageIdAndReaderIdentifierValue(message.metaData().messageId(),
					message.metaData().readerIdentifier().value());

			assertThat(result).isTrue();
		}

		@Test
		void shouldNotCloseMessageWithUnknownReaderIdentifier() {
			var result = xtaMessageRepository.closeByMessageIdAndReaderIdentifierValue(message.metaData().messageId(), "unknown");

			assertThat(result).isFalse();
		}

		@Test
		void shouldNotCloseMessageWithModifiedMessageId() {
			var result = xtaMessageRepository.closeByMessageIdAndReaderIdentifierValue(message.metaData().messageId() + "1",
					message.metaData().readerIdentifier().value());

			assertThat(result).isFalse();
		}
	}

	@DisplayName("close")
	@Nested
	class TestClose {
		@BeforeEach
		void beforeEach() {
			xtaMessageRepository = new XtaMessageRepository();
		}

		@Test
		void shouldSetClosedStatus() {
			xtaMessageRepository.add(message);

			xtaMessageRepository.close(message.metaData().messageId());
			var report = xtaMessageRepository.getTransportReport(message.metaData().messageId());

			assertThat(report).isPresent();
			assertThat(report.get().status()).isEqualTo(XtaTransportReport.MessageStatus.GRUEN);
		}

		@Test
		void shouldReturnTrueOnSuccess() {
			xtaMessageRepository.add(message);

			var result = xtaMessageRepository.close(message.metaData().messageId());

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalseOnMissingMessageId() {
			var result = xtaMessageRepository.close(message.metaData().messageId());

			assertThat(result).isFalse();
		}

	}

	@DisplayName("find open messages by reader identifier value limited to")
	@Nested
	class TestFindOpenMessagesByReaderIdentifierValueLimitedTo {

		@BeforeEach
		void beforeEach() {
			xtaMessageRepository.add(XtaMessageTestFactory.create("1"));
			xtaMessageRepository.add(XtaMessageTestFactory.create("2"));
			xtaMessageRepository.add(XtaMessageTestFactory.create("3"));
			xtaMessageRepository.add(XtaMessageTestFactory.create("4"));
			xtaMessageRepository.close("4");
		}

		@DisplayName("should return limit")
		@ParameterizedTest
		@ValueSource(ints = { 0, 1, 2 })
		void shouldReturnLimit(int limit) {
			var result = xtaMessageRepository.findOpenMessagesByReaderIdentifierValueLimitedTo(READER_IDENTIFIER.value(), limit);

			assertThat(result).hasSize(limit);
		}

		@DisplayName("should return all open messages")
		@ParameterizedTest
		@ValueSource(ints = { 3, 4, 5 })
		void shouldReturnAllOpenMessages(int limit) {
			var result = xtaMessageRepository.findOpenMessagesByReaderIdentifierValueLimitedTo(READER_IDENTIFIER.value(), limit);

			assertThat(result).hasSize(3);
		}

		@DisplayName("should not find any with unknown reader identifier")
		@Test
		void shouldNotFindAnyWithUnknownReaderIdentifier() {
			var result = xtaMessageRepository.findOpenMessagesByReaderIdentifierValueLimitedTo("unknown", 1);

			assertThat(result).isEmpty();
		}

	}

	@DisplayName("count open messages by reader identifier value")
	@Nested
	class TestCountOpenMessagesByReaderIdentifierValue {
		@BeforeEach
		void beforeEach() {
			doReturn(Stream.of(XtaMessageTestFactory.create("1"), XtaMessageTestFactory.create("2"),
					XtaMessageTestFactory.create("3"))).when(xtaMessageRepository).findOpenMessagesByReaderIdentifierValue(READER_IDENTIFIER.value());
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var count = xtaMessageRepository.countOpenMessagesByReaderIdentifierValue(READER_IDENTIFIER.value());

			assertThat(count).isEqualTo(3);
		}

	}

	@DisplayName("find open messages by reader identifier value")
	@Nested
	class TestFindOpenMessagesByReaderIdentifierValue {
		private static final String OTHER_READER_IDENTIFIER = "readerIdentifier";

		@BeforeEach
		void beforeEach() {
			xtaMessageRepository = new XtaMessageRepository();
			xtaMessageRepository.add(XtaMessageTestFactory.create("1"));
			xtaMessageRepository.add(XtaMessageTestFactory.create("2"));
			xtaMessageRepository.add(XtaMessageTestFactory.create("3"));
			xtaMessageRepository.add(XtaMessageTestFactory.create("4"));
			xtaMessageRepository.close("4");
			xtaMessageRepository.add(XtaMessageTestFactory.createBuilder()
					.metaData(XtaMessageMetaDataTestFactory.createBuilder()
							.readerIdentifier(XtaIdentifier.builder().value(OTHER_READER_IDENTIFIER).build())
							.build())
					.build());
		}

		@DisplayName("should return message for reader identifier")
		@Test
		void shouldReturnMessageForReaderIdentifier() {
			var result = xtaMessageRepository.findOpenMessagesByReaderIdentifierValue(READER_IDENTIFIER.value()).toList();

			assertThat(result).hasSize(3);
		}

		@DisplayName("should return messages for other reader identifier")
		@Test
		void shouldReturnMessagesForOtherReaderIdentifier() {
			var result = xtaMessageRepository.findOpenMessagesByReaderIdentifierValue(OTHER_READER_IDENTIFIER).toList();

			assertThat(result).hasSize(1);
		}

		@DisplayName("should return no messages for unknown reader identifier")
		@Test
		void shouldReturnNoMessagesForUnknownReaderIdentifier() {
			var result = xtaMessageRepository.findOpenMessagesByReaderIdentifierValue("unknown").toList();

			assertThat(result).isEmpty();
		}

	}
}
