package de.ozgcloud.xta.test.app.mapper;

import static de.ozgcloud.xta.test.app.mapper.ExceptionMapper.*;
import static org.assertj.core.api.Assertions.*;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;

import genv3.de.xoev.transport.xta.x211.InvalidMessageIDException;
import genv3.de.xoev.transport.xta.x211.ParameterIsNotValidException;
import genv3.de.xoev.transport.xta.x211.XTAWSTechnicalProblemException;

class ExceptionMapperTest {

	private final ExceptionMapper mapper = Mappers.getMapper(ExceptionMapper.class);

	@DisplayName("create invalid message id exception")
	@Nested
	class TestCreateInvalidMessageIdException {
		private static final String MESSAGE_ID = "messageid:test";
		private static final String IDENTIFIER = "Author";

		@DisplayName("should map message id and identifier to message")
		@Test
		void shouldMapMessage() {
			var result = createException();

			assertThat(result.getMessage()).isEqualTo("No message with messageID \"messageid:test\" available for identifier \"Author\".");
		}

		@DisplayName("should map error code")
		@Test
		void shouldMapErrorCode() {
			var result = createException().getFaultInfo().getErrorCode();

			assertThat(result).usingRecursiveComparison()
					.isEqualTo(mapper.mapCodeFehlernummer(ERROR_CODE_MESSAGE_ID_UNKNOWN, ERROR_CODE_MESSAGE_ID_UNKNOWN_DESCRIPTION));
		}

		private InvalidMessageIDException createException() {
			return mapper.createInvalidMessageIDException(MESSAGE_ID, IDENTIFIER);
		}
	}

	@DisplayName("create technical problem exception")
	@Nested
	class TestCreateTechnicalProblemException {
		private final String exceptionMessage = "test message";

		@DisplayName("should map message")
		@Test
		void shouldMapMessage() {
			var result = createException();

			assertThat(result.getMessage()).isEqualTo(exceptionMessage);
		}

		@DisplayName("should map error code")
		@Test
		void shouldMapErrorCode() {
			var result = createException().getFaultInfo().getErrorCode();

			assertThat(result).usingRecursiveComparison()
					.isEqualTo(mapper.mapCodeFehlernummer(ERROR_CODE_INTERNAL_TECHNICAL_PROBLEM, ERROR_CODE_INTERNAL_TECHNICAL_PROBLEM_DESCRIPTION));
		}

		private XTAWSTechnicalProblemException createException() {
			return mapper.createTechnicalProblemException(exceptionMessage);
		}

	}

	@DisplayName("create parameter is not valid exception")
	@Nested
	class TestCreateParameterIsNotValidException {
		private final String exceptionMessage = "test message";

		@DisplayName("should map message")
		@Test
		void shouldMapMessage() {
			var result = createException();

			assertThat(result.getMessage()).isEqualTo(exceptionMessage);
		}

		@DisplayName("should map error code")
		@Test
		void shouldMapErrorCode() {
			var result = createException().getFaultInfo().getErrorCode();

			assertThat(result).usingRecursiveComparison()
					.isEqualTo(mapper.mapCodeFehlernummer(ERROR_CODE_PARAMETER_MISSING, ERROR_CODE_PARAMETER_MISSING_DESCRIPTION));
		}

		private ParameterIsNotValidException createException() {
			return mapper.createParameterIsNotValidException(exceptionMessage);
		}
	}

	@DisplayName("map code error")
	@Nested
	class TestMapCodeError {
		private final String code = "test code";
		private final String name = "test name";

		@DisplayName("should map code")
		@Test
		void shouldMapCode() {
			var result = mapper.mapCodeFehlernummer(code, name);

			assertThat(result.getCode()).isEqualTo(code);
		}

		@DisplayName("should map name")
		@Test
		void shouldMapName() {
			var result = mapper.mapCodeFehlernummer(code, name);

			assertThat(result.getName()).isEqualTo(name);
		}

		@DisplayName("should map list uri")
		@Test
		void shouldMapListUri() {
			var result = mapper.mapCodeFehlernummer(code, name);

			assertThat(result.getListURI()).isEqualTo(ERROR_CODE_LIST_URN);
		}

		@DisplayName("should map list version id")
		@Test
		void shouldMapListVersionId() {
			var result = mapper.mapCodeFehlernummer(code, name);

			assertThat(result.getListVersionID()).isEqualTo(ERROR_CODE_LIST_VERSION);
		}
	}
}