package de.ozgcloud.xta.test.app.mapper;

import static de.ozgcloud.xta.test.app.factory.LookupServiceRequestTestFactory.*;
import static de.ozgcloud.xta.test.app.factory.MsgBoxCloseRequestTypeTestFactory.*;
import static de.ozgcloud.xta.test.app.factory.MsgBoxFetchRequestTestFactory.*;
import static de.ozgcloud.xta.test.app.factory.PartyIdentifierTypeTestFactory.*;
import static de.ozgcloud.xta.test.app.factory.XtaMessageMetaDataTestFactory.*;
import static org.assertj.core.api.Assertions.*;

import java.nio.file.Files;

import org.apache.cxf.ws.addressing.AttributedURIType;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;

import de.ozgcloud.xta.test.app.factory.AttributedURITypeTestFactory;
import de.ozgcloud.xta.test.app.factory.ContentTypeTestFactory;
import de.ozgcloud.xta.test.app.factory.GenericContentContainerTestFactory;
import de.ozgcloud.xta.test.app.factory.LookupServiceRequestTestFactory;
import de.ozgcloud.xta.test.app.factory.MessageMetaDataTestFactory;
import de.ozgcloud.xta.test.app.factory.MsgBoxCloseRequestTypeTestFactory;
import de.ozgcloud.xta.test.app.factory.MsgBoxFetchRequestTestFactory;
import de.ozgcloud.xta.test.app.factory.MsgBoxStatusListRequestTypeTestFactory;
import de.ozgcloud.xta.test.app.factory.PartyTypeTestFactory;
import de.ozgcloud.xta.test.app.factory.QualifierTypeBusinessScenarioTestFactory;
import de.ozgcloud.xta.test.app.model.*;

import genv3.de.xoev.transport.xta.x211.ContentType;
import genv3.de.xoev.transport.xta.x211.GenericContentContainer;
import genv3.eu.osci.ws.x2008.x05.transport.MsgBoxFetchRequest;
import genv3.eu.osci.ws.x2008.x05.transport.MsgSelector;
import genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData;
import genv3.eu.osci.ws.x2014.x10.transport.PartyIdentifierType;
import lombok.SneakyThrows;

public class RequestMapperTest {

	private final RequestMapper mapper = Mappers.getMapper(RequestMapper.class);

	@DisplayName("map XtaIdentifier")
	@Nested
	class TestMapXtaIdentifier {
		private static final PartyIdentifierType CLIENT_IDENTIFIER = PartyTypeTestFactory.create().getIdentifier();

		@DisplayName("should map name")
		@Test
		void shouldMapName() {
			var result = doMapping();

			assertThat(result.name()).isEqualTo(CLIENT_IDENTIFIER.getName());
		}

		@DisplayName("should map category")
		@Test
		void shouldMapCategory() {
			var result = doMapping();

			assertThat(result.category()).isEqualTo(CLIENT_IDENTIFIER.getCategory());
		}

		@DisplayName("should map value")
		@Test
		void shouldMapValue() {
			var result = doMapping();

			assertThat(result.value()).isEqualTo(CLIENT_IDENTIFIER.getValue());
		}

		private XtaIdentifier doMapping() {
			return mapper.mapXtaIdentifier(CLIENT_IDENTIFIER);
		}
	}

	@Nested
	class MapXtaMessageFromGenericContentContainer {
		private MessageMetaData messageMetaData;
		private GenericContentContainer genericContentContainer;
		private ContentType messageContentType;
		private ContentType attachmentContentType;

		@BeforeEach
		void beforeEach() {
			messageMetaData = MessageMetaDataTestFactory.create();
			genericContentContainer = GenericContentContainerTestFactory.create();
			messageContentType = ContentTypeTestFactory.createContentType();
			attachmentContentType = ContentTypeTestFactory.createContentType();
			var container = genericContentContainer.getContentContainer();
			container.setMessage(messageContentType);
			container.getAttachment().add(attachmentContentType);
		}

		@DisplayName("Should map message and attachment files from content container")
		@Test
		void shouldMapFiles() {
			var xtaMessage = mapper.mapXtaMessageFromGenericContentContainer(genericContentContainer, messageMetaData);
			var messageFile = xtaMessage.messageFile();
			var attachmentFiles = xtaMessage.attachmentFiles();

			assertThat(attachmentFiles).hasSize(1);
			assertXtaFileEqualsContentType(messageFile, messageContentType);
			assertXtaFileEqualsContentType(attachmentFiles.getFirst(), attachmentContentType);
		}

		@SneakyThrows
		void assertXtaFileEqualsContentType(XtaFile xtaFile, ContentType contentType) {
			assertThat(xtaFile.id()).isEqualTo(contentType.getId());
			assertThat(xtaFile.name()).isEqualTo(contentType.getFilename());
			assertThat(xtaFile.size()).isEqualTo(contentType.getSize());

			String content = Files.readString(xtaFile.file().toPath());
			assertThat(content).isEqualTo(ContentTypeTestFactory.CONTENT);
		}

		@DisplayName("Should map XtaMessageMetaData from MessageMetaData")
		@Test
		void shouldMapMetaData() {
			var xtaMessage = mapper.mapXtaMessageFromGenericContentContainer(genericContentContainer, messageMetaData);
			var xtaMessageMetaData = xtaMessage.metaData();

			assertThat(xtaMessageMetaData.service()).isEqualTo(MessageMetaDataTestFactory.QUALIFER_SERVICE);
			assertThat(xtaMessageMetaData.messageTypeCode()).isEqualTo(MessageMetaDataTestFactory.QUALIFER_MESSAGE_TYPE_CODE);
			assertThat(xtaMessageMetaData.messageTypePayloadSchema()).isEqualTo(MessageMetaDataTestFactory.QUALIFER_MESSAGE_TYPE_PAYLOAD_SCHEMA);
			assertThat(xtaMessageMetaData.messageId()).isEqualTo(MessageMetaDataTestFactory.MESSAGE_ID);
			assertThat(xtaMessageMetaData.messageSize()).isEqualTo(MessageMetaDataTestFactory.MESSAGE_SIZE);
		}

		@DisplayName("Should map author and reader identifiers from MessageMetaData")
		@Test
		void shouldMapIdentifier() {
			var xtaMessage = mapper.mapXtaMessageFromGenericContentContainer(genericContentContainer, messageMetaData);
			var authorIdentifier = xtaMessage.metaData().authorIdentifier();
			var readerIdentifier = xtaMessage.metaData().readerIdentifier();

			assertIdentifiersAreEqual(authorIdentifier, AUTHOR_PARTY_IDENTIFIER);
			assertIdentifiersAreEqual(readerIdentifier, READER_PARTY_IDENTIFIER);
		}

		void assertIdentifiersAreEqual(XtaIdentifier identifier, PartyIdentifierType identifierType) {
			assertThat(identifier.name()).isEqualTo(identifierType.getName());
			assertThat(identifier.category()).isEqualTo(identifierType.getCategory());
			assertThat(identifier.value()).isEqualTo(identifierType.getValue());
		}

		@DisplayName("Should map a defined business scenario from MessageMetaData")
		@Test
		void shouldMapDefinedBusinessScenario() {
			var businessScenario = QualifierTypeBusinessScenarioTestFactory.createDefinedBusinessScenario();
			messageMetaData.getQualifier().setBusinessScenario(businessScenario);

			var xtaMessageMetaData = mapper.mapXtaMessageMetaData(messageMetaData);

			assertThat(xtaMessageMetaData.businessScenarioCode()).isEqualTo(QualifierTypeBusinessScenarioTestFactory.BUSINESS_SCENARIO_CODE);
			assertThat(xtaMessageMetaData.businessScenarioName()).isEqualTo(QualifierTypeBusinessScenarioTestFactory.BUSINESS_SCENARIO_NAME);
			assertThat(xtaMessageMetaData.businessScenarioListUri()).isEqualTo(QualifierTypeBusinessScenarioTestFactory.BUSINESS_SCENARIO_LIST_URI);
			assertThat(xtaMessageMetaData.businessScenarioListVersionId()).isEqualTo(
					QualifierTypeBusinessScenarioTestFactory.BUSINESS_SCENARIO_LIST_VERSION_ID);
		}

		@DisplayName("Should map an undefined business scenario from MessageMetaData")
		@Test
		void shouldMapUndefinedBusinessScenario() {
			var businessScenario = QualifierTypeBusinessScenarioTestFactory.createUndefinedBusinessScenario();
			messageMetaData.getQualifier().setBusinessScenario(businessScenario);

			var xtaMessageMetaData = mapper.mapXtaMessageMetaData(messageMetaData);

			assertThat(xtaMessageMetaData.businessScenarioCode()).isEqualTo(QualifierTypeBusinessScenarioTestFactory.BUSINESS_SCENARIO_CODE);
			assertThat(xtaMessageMetaData.businessScenarioListUri()).isNull();
		}
	}

	@Nested
	class MapMessageIdFromMsgBoxFetchRequest {
		public static final String MESSAGE_ID = "de:xta:messageid:xta-test-server:1234";

		private MsgBoxFetchRequest fetchRequest;

		@BeforeEach
		void beforeEach() {
			AttributedURIType messageID = AttributedURITypeTestFactory.create(MESSAGE_ID);

			var msgSelector = new MsgSelector();
			msgSelector.getMessageID().add(messageID);

			fetchRequest = new MsgBoxFetchRequest();
			fetchRequest.setMsgSelector(msgSelector);
		}

		@DisplayName("Should map message id from MsgBoxRequest")
		@Test
		void shouldMapMessageId() {
			var messageId = mapper.mapMessageIdFromMsgBoxFetchRequest(fetchRequest);

			assertThat(messageId).isEqualTo(MESSAGE_ID);
		}
	}

	@DisplayName("map message meta data listing request")
	@Nested
	class TestMapMessageMetaDataListingRequest {
		private static final int MAX_LIST_ITEMS = 10;

		@DisplayName("should map client identifier")
		@Test
		void shouldMapClientIdentifier() {
			var result = doMapping();

			assertThat(result.clientIdentifier())
					.usingRecursiveComparison()
					.isEqualTo(READER_IDENTIFIER);
		}

		@DisplayName("should map max list items")
		@Test
		void shouldMaxMaxListItems() {
			var result = doMapping();

			assertThat(result.maxListItems()).isEqualTo(MAX_LIST_ITEMS);
		}

		private XtaMessageMetaDataListingRequest doMapping() {
			return mapper.mapMessageMetaDataListingRequest(
					MsgBoxStatusListRequestTypeTestFactory.create(MAX_LIST_ITEMS),
					READER_PARTY_TYPE
			);
		}
	}

	@DisplayName("map xta close request")
	@Nested
	class TestMapXtaCloseRequest {
		@DisplayName("should map message id")
		@Test
		void shouldMapMessageId() {
			var result = doMapping();

			assertThat(result.messageId()).isEqualTo(CLOSE_REQUEST_MESSAGE_ID);
		}

		@DisplayName("should map client identifier")
		@Test
		void shouldMapClientIdentifier() {
			var result = doMapping();

			assertThat(result.clientIdentifier())
					.usingRecursiveComparison()
					.isEqualTo(READER_IDENTIFIER);
		}

		private XtaCloseRequest doMapping() {
			return mapper.mapXtaCloseRequest(MsgBoxCloseRequestTypeTestFactory.create(),
					PartyTypeTestFactory.create());
		}
	}

	@DisplayName("map xta get message request")
	@Nested
	class TestMapXtaGetMessageRequest {
		@DisplayName("should map message id")
		@Test
		void shouldMapMessageId() {
			var result = doMapping();

			assertThat(result.messageId()).isEqualTo(GET_MESSAGE_REQUEST_MESSAGE_ID);
		}

		@DisplayName("should map client identifier")
		@Test
		void shouldMapClientIdentifier() {
			var result = doMapping();

			assertThat(result.clientIdentifier())
					.usingRecursiveComparison()
					.isEqualTo(READER_IDENTIFIER);
		}

		private XtaGetMessageRequest doMapping() {
			return mapper.mapXtaGetMessageRequest(MsgBoxFetchRequestTestFactory.create(),
					PartyTypeTestFactory.create());
		}
	}

	@DisplayName("map xta lookup service request")
	@Nested
	class TestMapXtaLookupServiceRequest {

		@DisplayName("should map service")
		@Test
		void shouldMapService() {
			var result = doMapping();

			assertThat(result.service()).isEqualTo(SERVICE_TYPE);
		}

		@DisplayName("should map reader identifier")
		@Test
		void shouldMapReaderIdentifier() {
			var result = doMapping();

			assertThat(result.readerIdentifier())
					.usingRecursiveComparison()
					.isEqualTo(READER_IDENTIFIER);
		}

		@DisplayName("should map author identifier value")
		@Test
		void shouldMapAuthorIdentifierValue() {
			var result = doMapping();

			assertThat(result.authorIdentifier())
					.usingRecursiveComparison()
					.isEqualTo(AUTHOR_IDENTIFIER);
		}

		private XtaLookupServiceRequest doMapping() {
			return mapper.mapXtaLookupServiceRequest(LookupServiceRequestTestFactory.create(),
					PartyTypeTestFactory.createWithAuthor());
		}
	}

	@DisplayName("map xta get transport report request")
	@Nested
	class TestMapXtaGetTransportReportRequest {
		public static final String MESSAGE_ID = "urn:de:xta:messageid:1";

		@DisplayName("should map message id")
		@Test
		void shouldMapMessageId() {
			var result = doMapping();

			assertThat(result.messageId()).isEqualTo(MESSAGE_ID);
		}

		@DisplayName("should map client identifier")
		@Test
		void shouldMapReaderIdentifier() {
			var result = doMapping();

			assertThat(result.clientIdentifier())
					.usingRecursiveComparison()
					.isEqualTo(READER_IDENTIFIER);
		}

		private XtaGetTransportReportRequest doMapping() {
			return mapper.mapXtaGetTransportReportRequest(AttributedURITypeTestFactory.create(MESSAGE_ID),
					PartyTypeTestFactory.create());
		}
	}
}
