package de.ozgcloud.xta.test.app.mapper;

import static de.ozgcloud.xta.test.app.factory.MessageMetaDataTestFactory.MESSAGE_SIZE;
import static de.ozgcloud.xta.test.app.factory.PartyIdentifierTypeTestFactory.READER_PARTY_IDENTIFIER;
import static de.ozgcloud.xta.test.app.factory.XtaLookupServiceRequestTestFactory.LOOKUP_SERVICE_SERVICE;
import static de.ozgcloud.xta.test.app.factory.XtaMessageMetaDataTestFactory.*;
import static de.ozgcloud.xta.test.app.mapper.ResponseMapper.*;
import static org.assertj.core.api.Assertions.assertThat;

import java.math.BigInteger;
import java.nio.charset.StandardCharsets;
import java.time.ZonedDateTime;
import java.util.List;

import genv3.de.xoev.transport.xta.x211.ContentType;
import genv3.de.xoev.transport.xta.x211.LookupServiceResponse;
import genv3.de.xoev.transport.xta.x211.LookupServiceResultType;
import genv3.de.xoev.transport.xta.x211.TransportReport;
import genv3.eu.osci.ws.x2008.x05.transport.MsgStatusListType;
import genv3.eu.osci.ws.x2014.x10.transport.DeliveryAttributesType;
import genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData;
import genv3.eu.osci.ws.x2014.x10.transport.QualifierType;
import lombok.SneakyThrows;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mapstruct.factory.Mappers;

import de.ozgcloud.xta.test.app.factory.XtaFileTestFactory;
import de.ozgcloud.xta.test.app.factory.XtaLookupServiceRequestTestFactory;
import de.ozgcloud.xta.test.app.factory.XtaMessageMetaDataTestFactory;
import de.ozgcloud.xta.test.app.factory.XtaMessageTestFactory;
import de.ozgcloud.xta.test.app.model.XtaFile;
import de.ozgcloud.xta.test.app.model.XtaMessageMetaData;
import de.ozgcloud.xta.test.app.model.XtaMessageMetaDataListing;
import de.ozgcloud.xta.test.app.model.XtaTransportReport;

public class ResponseMapperTest {

	private final ResponseMapper mapper = Mappers.getMapper(ResponseMapper.class);

	@DisplayName("map PartyIdentifierType")
	@Nested
	class TestMapPartyIdentifierType {

		@DisplayName("should map type")
		@Test
		void shouldMapType() {
			var result = mapper.mapPartyIdentifierType(AUTHOR_IDENTIFIER);

			assertThat(result.getType()).isEqualTo(IDENTIFIER_TYPE);
		}

		@DisplayName("should map value")
		@Test
		void shouldMapValue() {
			var result = mapper.mapPartyIdentifierType(AUTHOR_IDENTIFIER);

			assertThat(result.getValue()).isEqualTo(AUTHOR_IDENTIFIER.value());
		}

		@DisplayName("should map category")
		@Test
		void shouldMapCategory() {
			var result = mapper.mapPartyIdentifierType(AUTHOR_IDENTIFIER);

			assertThat(result.getCategory()).isEqualTo(AUTHOR_IDENTIFIER.category());
		}

		@DisplayName("should map name")
		@Test
		void shouldMapName() {
			var result = mapper.mapPartyIdentifierType(AUTHOR_IDENTIFIER);

			assertThat(result.getName()).isEqualTo(AUTHOR_IDENTIFIER.name());
		}
	}

	@Nested
	class MapGenericContentContainerFromXtaMessage {
		private XtaFile messageXtaFile;
		private XtaFile attachmentXtaFile;

		@BeforeEach
		void beforeEach() {
			messageXtaFile = XtaFileTestFactory.create();
			attachmentXtaFile = XtaFileTestFactory.create();
		}

		@DisplayName("should map generic content container from xta message")
		@Test
		void shouldMap() {
			var xtaMessage = XtaMessageTestFactory.createBuilder("messageId")
					.messageFile(messageXtaFile)
					.attachmentFiles(List.of(attachmentXtaFile))
					.build();

			var genericContentContainer = mapper.mapGenericContentContainerFromXtaMessage(xtaMessage);
			var message = genericContentContainer.getContentContainer().getMessage();
			var attachmentList = genericContentContainer.getContentContainer().getAttachment();

			assertThat(attachmentList).hasSize(1);

			assertContentTypeEqualsXtaFile(message, messageXtaFile);
			assertContentTypeEqualsXtaFile(attachmentList.getFirst(), attachmentXtaFile);
		}

		@SneakyThrows
		void assertContentTypeEqualsXtaFile(ContentType contentType, XtaFile xtaFile) {
			assertThat(contentType.getId()).isEqualTo(xtaFile.id());
			assertThat(contentType.getFilename()).isEqualTo(xtaFile.name());
			assertThat(contentType.getSize()).isEqualTo(xtaFile.size());

			String content = new String(contentType.getValue().getInputStream().readAllBytes(), StandardCharsets.UTF_8);
			assertThat(content).isEqualTo(XtaFileTestFactory.XTA_CONTENT);
		}
	}

	@DisplayName("map MsgBoxResponseType error")
	@Nested
	class TestMapMsgBoxResponseTypeError {
		private static final String ERROR_STRING = "some error";

		@DisplayName("should map no message available reason")
		@Test
		void shouldMapNoMessageAvailableReason() {
			var result = mapper.mapMsgBoxResponseTypeError(ERROR_STRING);

			assertThat(result.getNoMessageAvailable().getReason()).isEqualTo(ERROR_STRING);
		}

		@DisplayName("should map request id")
		@Test
		void shouldMapRequestId() {
			var result = mapper.mapMsgBoxResponseTypeError(ERROR_STRING);

			assertThat(result.getMsgBoxRequestID()).isEqualTo(ResponseMapper.REQUEST_ID);
		}
	}

	@DisplayName("map MsgStatusListType")
	@Nested
	class TestMapMsgStatusListType {
		private XtaMessageMetaDataListing xtaMessageMetaDataListing;

		@BeforeEach
		void beforeEach() {
			xtaMessageMetaDataListing = XtaMessageMetaDataListing.builder()
					.pendingMessageCount(BigInteger.ONE)
					.messages(List.of(XtaMessageMetaDataTestFactory.create()))
					.build();
		}

		@DisplayName("should map message meta data")
		@Test
		void shouldMapMessageMetaData() {
			var result = mapper.mapMsgStatusListType(xtaMessageMetaDataListing);

			assertThat(result.getMessageMetaData()).hasSize(1);
		}

		@DisplayName("should map message size")
		@Test
		void shouldMapMessageSize() {
			var result = mapFirstMessageMetaData();

			assertThat(result.getMsgSize()).isEqualTo(MESSAGE_SIZE);
		}

		@DisplayName("should map service")
		@Test
		void shouldMapService() {
			var result = mapFirstMessageMetaData();

			assertThat(result.getQualifier().getService()).isEqualTo(SERVICE);
		}

		@DisplayName("with defined business scenario")
		@Nested
		class TestWithDefinedBusinessScenario {

			@BeforeEach
			void beforeEach() {
				xtaMessageMetaDataListing = XtaMessageMetaDataListing.builder()
						.pendingMessageCount(BigInteger.ONE)
						.messages(List.of(XtaMessageMetaDataTestFactory.create()))
						.build();
			}

			@DisplayName("should map code")
			@Test
			void shouldMapCode() {
				var result = mapFirstMessageMetaDataBusinessScenario();

				assertThat(result.getDefined().getCode()).isEqualTo(BUSINESS_SCENARIO_CODE);
			}

			@DisplayName("should map name")
			@Test
			void shouldMapName() {
				var result = mapFirstMessageMetaDataBusinessScenario();

				assertThat(result.getDefined().getName()).isEqualTo(BUSINESS_SCENARIO_NAME);
			}

			@DisplayName("should map list uri")
			@Test
			void shouldMapListUri() {
				var result = mapFirstMessageMetaDataBusinessScenario();

				assertThat(result.getDefined().getListURI()).isEqualTo(BUSINESS_SCENARIO_LIST_URI);
			}

			@DisplayName("should map list version id")
			@Test
			void shouldMapListVersionId() {
				var result = mapFirstMessageMetaDataBusinessScenario();

				assertThat(result.getDefined().getListVersionID()).isEqualTo(BUSINESS_SCENARIO_LIST_VERSION_ID);
			}
		}

		@DisplayName("with undefined business scenario")
		@Nested
		class TestWithUndefinedBusinessScenario {
			@BeforeEach
			void mock() {
				xtaMessageMetaDataListing = XtaMessageMetaDataListing.builder()
						.pendingMessageCount(BigInteger.ONE)
						.messages(List.of(XtaMessageMetaDataTestFactory.createBuilder()
								.businessScenarioListUri(null)
								.build()))
						.build();

			}

			@DisplayName("should map code")
			@Test
			void shouldMapCode() {
				var result = mapFirstMessageMetaDataBusinessScenario();

				assertThat(result.getUndefined()).isEqualTo(BUSINESS_SCENARIO_CODE);
			}

			@DisplayName("should map defined null")
			@Test
			void shouldMapDefinedNull() {
				var result = mapFirstMessageMetaDataBusinessScenario();

				assertThat(result.getDefined()).isNull();
			}
		}

		@DisplayName("should map message type code")
		@Test
		void shouldMapMessageTypeCode() {
			var result = mapFirstMessageMetaData();

			assertThat(result.getQualifier().getMessageType().getCode()).isEqualTo(MESSAGE_TYPE_CODE);
		}

		@DisplayName("should map message type list uri")
		@Test
		void shouldMapMessageTypeListUri() {
			var result = mapFirstMessageMetaData();

			assertThat(result.getQualifier().getMessageType().getListURI()).isEqualTo(MESSAGE_TYPE_LIST_URI);
		}

		@DisplayName("should map message type list version id")
		@Test
		void shouldMapMessageTypeListVersionId() {
			var result = mapFirstMessageMetaData();

			assertThat(result.getQualifier().getMessageType().getListVersionID()).isEqualTo(MESSAGE_TYPE_LIST_VERSION_ID);
		}

		@DisplayName("should map author identifier")
		@Test
		void shouldMapAuthorIdentifier() {
			var result = mapFirstMessageMetaData();

			assertThat(result.getOriginators().getAuthor().getIdentifier())
					.usingRecursiveComparison()
					.isEqualTo(mapper.mapPartyIdentifierType(AUTHOR_IDENTIFIER));
		}

		@DisplayName("should set default sender identifier")
		@Test
		void shouldSetDefaultSenderIdentifier() {
			var result = mapFirstMessageMetaData();

			assertThat(result.getOriginators().getSender().getIdentifier())
					.usingRecursiveComparison()
					.isEqualTo(mapper.mapPartyIdentifierType(SENDER_IDENTIFIER));
		}

		@DisplayName("should map reader identifier")
		@Test
		void shouldMapReaderIdentifier() {
			var result = mapFirstMessageMetaData();

			assertThat(result.getDestinations().getReader().getIdentifier())
					.usingRecursiveComparison()
					.isEqualTo(mapper.mapPartyIdentifierType(READER_IDENTIFIER));
		}

		@DisplayName("should set default delivery attribute delivery")
		@Test
		void shouldSetDefaultDeliveryAttributeDelivery() {
			var result = mapFirstMessageMetaDataDelivery();

			assertThat(result.getDelivery()).isNotNull();
		}

		@DisplayName("should set default delivery attribute initial fetch")
		@Test
		void shouldSetDefaultDeliveryAttributeInitialFetch() {
			var result = mapFirstMessageMetaDataDelivery();

			assertThat(result.getInitialFetch()).isNotNull();
		}

		@DisplayName("should set default delivery attribute initial send")
		@Test
		void shouldSetDefaultDeliveryAttributeInitialSend() {
			var result = mapFirstMessageMetaDataDelivery();

			assertThat(result.getInitialSend()).isNotNull();
		}

		@DisplayName("should set default delivery attribute origin")
		@Test
		void shouldSetDefaultDeliveryAttributeOrigin() {
			var result = mapFirstMessageMetaDataDelivery();

			assertThat(result.getOrigin()).isNotNull();
		}

		private MsgStatusListType doMapping() {
			return mapper.mapMsgStatusListType(xtaMessageMetaDataListing);
		}

		private DeliveryAttributesType mapFirstMessageMetaDataDelivery() {
			return mapFirstMessageMetaData().getDeliveryAttributes();
		}

		private MessageMetaData mapFirstMessageMetaData() {
			return doMapping().getMessageMetaData().getFirst();
		}

		private QualifierType.BusinessScenario mapFirstMessageMetaDataBusinessScenario() {
			return mapFirstMessageMetaData().getQualifier().getBusinessScenario();
		}
	}

	@DisplayName("map MsgBoxResponseType")
	@Nested
	class TestMapMsgBoxResponseType {
		private XtaMessageMetaDataListing xtaMessageMetaDataListing;

		@DisplayName("with no pending messages")
		@Nested
		class TestWithNoPendingMessages {
			@BeforeEach
			void beforeEach() {
				xtaMessageMetaDataListing = XtaMessageMetaDataListing.builder()
						.pendingMessageCount(BigInteger.ZERO)
						.messages(List.of())
						.build();
			}

			@DisplayName("should set no message available reason")
			@Test
			void shouldSetNoMessageAvailableReason() {
				var msgBoxResponseType = mapper.mapMsgBoxResponseType(xtaMessageMetaDataListing);

				assertThat(msgBoxResponseType.getNoMessageAvailable().getReason()).isEqualTo(ResponseMapper.NO_MESSAGES_REASON);
			}

			@DisplayName("should set items pending to null")
			@Test
			void shouldSetItemsPendingToNull() {
				var msgBoxResponseType = mapper.mapMsgBoxResponseType(xtaMessageMetaDataListing);

				assertThat(msgBoxResponseType.getItemsPending()).isNull();
			}
		}

		@DisplayName("with pending messages")
		@Nested
		class TestWithPendingMessages {
			@BeforeEach
			void beforeEach() {
				xtaMessageMetaDataListing = XtaMessageMetaDataListing.builder()
						.pendingMessageCount(BigInteger.ONE)
						.messages(List.of(XtaMessageMetaDataTestFactory.create()))
						.build();
			}

			@DisplayName("should set items pending")
			@Test
			void shouldSetItemsPending() {
				var msgBoxResponseType = mapper.mapMsgBoxResponseType(xtaMessageMetaDataListing);

				assertThat(msgBoxResponseType.getItemsPending()).isEqualTo(BigInteger.ONE);
			}

			@DisplayName("should set no message available to null")
			@Test
			void shouldSetNoMessageAvailableToNull() {
				var msgBoxResponseType = mapper.mapMsgBoxResponseType(xtaMessageMetaDataListing);

				assertThat(msgBoxResponseType.getNoMessageAvailable()).isNull();
			}
		}
	}

	@DisplayName("map MsgBoxResponseType from requestId")
	@Nested
	class TestMapMsgBoxResponseTypeFromRequestId {

		@DisplayName("should map message box response type from request id")
		@Test
		void shouldMapMsgBoxResponseTypeFromRequestId() {
			var result = mapper.mapMsgBoxResponseTypeFromRequestId(REQUEST_ID);

			assertThat(result.getMsgBoxRequestID()).isEqualTo(REQUEST_ID);
		}

		@DisplayName("should map items pending to zero")
		@Test
		void shouldMapItemsPendingToZero() {
			var result = mapper.mapMsgBoxResponseTypeFromRequestId(REQUEST_ID);

			assertThat(result.getItemsPending()).isEqualTo(BigInteger.ZERO);
		}

		@DisplayName("should set no messages available to null")
		@Test
		void shouldSetNoMessagesAvailableToNull() {
			var result = mapper.mapMsgBoxResponseTypeFromRequestId(REQUEST_ID);

			assertThat(result.getNoMessageAvailable()).isNull();
		}
	}

	@DisplayName("map lookup service response")
	@Nested
	class TestMapLookupServiceResponse {

		@DisplayName("should map service type")
		@Test
		void shouldMapServiceType() {
			var result = getFirstLookupServiceResult();

			assertThat(result.getServiceType()).isEqualTo(LOOKUP_SERVICE_SERVICE);
		}

		@DisplayName("should map reader identifier")
		@Test
		void shouldMapReaderIdentifier() {
			var result = getFirstLookupServiceResult();

			assertThat(result.getReader().getIdentifier())
					.usingRecursiveComparison()
					.isEqualTo(READER_PARTY_IDENTIFIER);
		}

		@DisplayName("should map service is available value")
		@Test
		void shouldMapServiceIsAvailableValue() {
			var result = getFirstLookupServiceResult();

			assertThat(result.getIsServiceAvailableValue().isServiceIsAvailable()).isTrue();
		}

		private LookupServiceResultType getFirstLookupServiceResult() {
			return doMapping().getLookupServiceResultList().getLookupServiceResult().getFirst();
		}

		private LookupServiceResponse doMapping() {
			return mapper.mapLookupServiceResponse(XtaLookupServiceRequestTestFactory.create());
		}
	}

	@DisplayName("map transport report")
	@Nested
	class TestMapTransportReport {
		private XtaMessageMetaData xtaMessageMetaData;
		private ZonedDateTime reportTime;

		@BeforeEach
		void beforeEach() {
			xtaMessageMetaData = XtaMessageMetaDataTestFactory.create();
			reportTime = ZonedDateTime.now();
		}

		@DisplayName("should map message meta data")
		@Test
		void shouldMapMessageMetaData() {
			var result = doMapping();

			assertThat(result.getMessageMetaData().getMsgIdentification().getMessageID().getValue()).isEqualTo(xtaMessageMetaData.messageId());
		}

		@DisplayName("should map message status")
		@ParameterizedTest
		@EnumSource(XtaTransportReport.MessageStatus.class)
		void shouldMapMessageStatus(XtaTransportReport.MessageStatus status) {
			var result = doMapping(status);

			assertThat(result.getMessageStatus().getStatus()).isEqualTo(status.ordinal());
		}

		@DisplayName("should map report time")
		@Test
		void shouldMapReportTime() {
			var result = doMapping();

			assertThat(result.getReportTime()).isEqualTo(mapper.mapXmlGregorianCalendarFromZonedDateTime(reportTime));
		}

		@DisplayName("should set server identity")
		@Test
		void shouldSetServerIdentity() {
			var result = doMapping();

			assertThat(result.getXTAServerIdentity()).isEqualTo("xta-test-server");
		}

		private TransportReport doMapping() {
			return doMapping(XtaTransportReport.MessageStatus.GRUEN);
		}

		private TransportReport doMapping(XtaTransportReport.MessageStatus status) {
			return mapper.mapTransportReport(XtaTransportReport.builder()
					.messageMetaData(xtaMessageMetaData)
					.status(status)
					.reportTime(reportTime)
					.build());
		}
	}

	@DisplayName("map XmlGregorianCalendar from LocalDateTime")
	@Nested
	class TestMapXmlGregorianCalendarFromLocalDateTime {
		private ZonedDateTime zonedDateTime;
		private static final String ISO_TIME_STRING = "2007-12-03T10:15:30.000Z";

		@BeforeEach
		void beforeEach() {
			zonedDateTime = ZonedDateTime.parse(ISO_TIME_STRING);
		}

		@DisplayName("should map xml gregorian calendar from zoned date time")
		@Test
		void shouldMapXmlGregorianCalendarFromLocalDateTime() {
			var result = mapper.mapXmlGregorianCalendarFromZonedDateTime(zonedDateTime);

			assertThat(result.toXMLFormat()).isEqualTo(ISO_TIME_STRING);
		}
	}
}
