package de.ozgcloud.xta.test.app.server;

import static de.ozgcloud.xta.test.app.factory.MessageMetaDataTestFactory.MESSAGE_ID;
import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;
import static org.mockito.Mockito.*;

import java.util.Optional;

import genv3.de.xoev.transport.xta.x211.InvalidMessageIDException;
import genv3.de.xoev.transport.xta.x211.LookupServiceRequest;
import genv3.de.xoev.transport.xta.x211.LookupServiceResponse;
import genv3.de.xoev.transport.xta.x211.TransportReport;
import genv3.eu.osci.ws.x2014.x10.transport.PartyType;
import lombok.SneakyThrows;
import org.apache.cxf.ws.addressing.AttributedURIType;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.xta.test.app.factory.XtaMessageMetaDataTestFactory;
import de.ozgcloud.xta.test.app.mapper.ExceptionMapper;
import de.ozgcloud.xta.test.app.mapper.RequestMapper;
import de.ozgcloud.xta.test.app.mapper.ResponseMapper;
import de.ozgcloud.xta.test.app.model.XtaGetTransportReportRequest;
import de.ozgcloud.xta.test.app.model.XtaLookupServiceRequest;
import de.ozgcloud.xta.test.app.model.XtaTransportReport;
import de.ozgcloud.xta.test.app.service.XtaMessageService;
import de.ozgcloud.xta.test.app.validation.RequestValidator;

public class ManagementPortImplTest {

	@InjectMocks
	private ManagementPortImpl managementPortImpl;

	@Mock
	private XtaMessageService xtaMessageService;
	@Mock
	private ExceptionMapper exceptionMapper;
	@Mock
	private ResponseMapper responseMapper;
	@Mock
	private RequestMapper requestMapper;
	@Mock
	private RequestValidator requestValidator;

	@DisplayName("create message id")
	@Nested
	class TestCreateMessageId {
		@Mock
		private PartyType authorIdentifier;

		@Mock
		private AttributedURIType createdMessageId;

		@BeforeEach
		void beforeEach() {
			when(xtaMessageService.createMessageId()).thenReturn(MESSAGE_ID);
			when(responseMapper.mapCreateMessageIdResponse(MESSAGE_ID)).thenReturn(createdMessageId);
		}

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldCreateMessageId() {
			var result = managementPortImpl.createMessageId(authorIdentifier);

			assertThat(result).isEqualTo(createdMessageId);
		}
	}

	@DisplayName("check account active")
	@Nested
	class TestCheckAccountActive {
		@Mock
		private PartyType authorIdentifier;

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldAcceptRequest() {
			managementPortImpl.checkAccountActive(authorIdentifier);
		}
	}

	@DisplayName("lookup service")
	@Nested
	class TestLookupService {
		@Mock
		private LookupServiceRequest lookupServiceRequest;
		@Mock
		private PartyType authorIdentifier;

		@Mock
		private XtaLookupServiceRequest request;

		@Mock
		private LookupServiceResponse response;

		@BeforeEach
		void beforeEach() {
			when(requestMapper.mapXtaLookupServiceRequest(lookupServiceRequest, authorIdentifier)).thenReturn(request);
			when(responseMapper.mapLookupServiceResponse(request)).thenReturn(response);
		}

		@DisplayName("should validate request")
		@Test
		@SneakyThrows
		void shouldValidateRequest() {
			managementPortImpl.lookupService(lookupServiceRequest, authorIdentifier);

			verify(requestValidator).validate(request);
		}

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			var result = managementPortImpl.lookupService(lookupServiceRequest, authorIdentifier);

			assertThat(result).isEqualTo(response);
		}
	}

	@DisplayName("get transport report")
	@Nested
	class TestGetTransportReport {
		private static final String MESSAGE_ID = "urn:de:xta:messageid:1";

		@Mock
		private AttributedURIType messageId;
		@Mock
		private PartyType authorIdentifier;
		@Mock
		private XtaTransportReport xtaTransportReport;
		@Mock
		private TransportReport transportReport;

		private XtaGetTransportReportRequest request;

		@BeforeEach
		void beforeEach() {
			request = XtaGetTransportReportRequest.builder()
					.messageId(MESSAGE_ID)
					.clientIdentifier(XtaMessageMetaDataTestFactory.READER_IDENTIFIER)
					.build();

			when(requestMapper.mapXtaGetTransportReportRequest(messageId, authorIdentifier)).thenReturn(request);
		}

		@DisplayName("with valid request")
		@Nested
		class TestWithValidRequest {
			@BeforeEach
			void beforeEach() {
				when(xtaMessageService.getTransportReport(request)).thenReturn(Optional.of(xtaTransportReport));
				when(responseMapper.mapTransportReport(xtaTransportReport)).thenReturn(transportReport);
			}

			@DisplayName("should return")
			@Test
			void shouldReturn() {
				var result = getTransportReport();

				assertThat(result).isEqualTo(transportReport);
			}
		}

		@DisplayName("with Invalid message id")
		@Nested
		class TestWithInvalidMessageId {
			@Mock
			private InvalidMessageIDException exception;

			@DisplayName("should throw exception")
			@Test
			void shouldThrowException() {
				when(xtaMessageService.getTransportReport(request)).thenReturn(Optional.empty());
				when(exceptionMapper.createInvalidMessageIDException(any(), any())).thenReturn(exception);

				assertThatThrownBy(TestGetTransportReport.this::getTransportReport)
						.isEqualTo(exception);
			}
		}

		@SneakyThrows
		private TransportReport getTransportReport() {
			return managementPortImpl.getTransportReport(messageId, authorIdentifier);
		}
	}
}