package de.ozgcloud.xta.test.app.server;

import static de.ozgcloud.xta.test.app.factory.MessageMetaDataTestFactory.*;
import static de.ozgcloud.xta.test.app.factory.XtaMessageMetaDataTestFactory.*;
import static de.ozgcloud.xta.test.app.mapper.ResponseMapper.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;

import jakarta.xml.ws.Holder;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.xta.test.app.mapper.ExceptionMapper;
import de.ozgcloud.xta.test.app.mapper.RequestMapper;
import de.ozgcloud.xta.test.app.mapper.ResponseMapper;
import de.ozgcloud.xta.test.app.model.XtaCloseRequest;
import de.ozgcloud.xta.test.app.model.XtaGetMessageRequest;
import de.ozgcloud.xta.test.app.model.XtaMessage;
import de.ozgcloud.xta.test.app.model.XtaMessageMetaDataListing;
import de.ozgcloud.xta.test.app.model.XtaMessageMetaDataListingRequest;
import de.ozgcloud.xta.test.app.service.XtaMessageService;
import de.ozgcloud.xta.test.app.validation.RequestValidator;
import genv3.de.xoev.transport.xta.x211.GenericContentContainer;
import genv3.de.xoev.transport.xta.x211.InvalidMessageIDException;
import genv3.de.xoev.transport.xta.x211.XTAWSTechnicalProblemException;
import genv3.eu.osci.ws.x2008.x05.transport.MsgBoxCloseRequestType;
import genv3.eu.osci.ws.x2008.x05.transport.MsgBoxFetchRequest;
import genv3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType;
import genv3.eu.osci.ws.x2008.x05.transport.MsgBoxStatusListRequestType;
import genv3.eu.osci.ws.x2008.x05.transport.MsgStatusListType;
import genv3.eu.osci.ws.x2014.x10.transport.MessageMetaData;
import genv3.eu.osci.ws.x2014.x10.transport.PartyType;
import lombok.SneakyThrows;

public class MsgBoxPortImplTest {
	@InjectMocks
	private MsgBoxPortImpl msgBoxPortImpl;

	@Mock
	private RequestMapper requestMapper;

	@Mock
	private RequestValidator requestValidator;

	@Mock
	private ExceptionMapper exceptionMapper;

	@Mock
	private ResponseMapper responseMapper;

	@Mock
	private XtaMessageService xtaMessageService;

	@DisplayName("get message")
	@Nested
	class TestGetMessage {
		private static final String MESSAGE_ID = "message-id";

		@Mock
		private MsgBoxFetchRequest fetchRequest;
		@Mock
		private PartyType authorIdentifier;
		@Mock
		private XtaMessage xtaMessage;
		@Mock
		private GenericContentContainer mappedGenericContentContainer;
		@Mock
		private MessageMetaData mappedMessageMetaData;
		@Mock
		private MsgBoxResponseType mappedFetchResponseHeader;

		private Holder<MessageMetaData> messageMetaDataHolder;
		private Holder<MsgBoxResponseType> responseHeaderHolder;
		private XtaGetMessageRequest request;

		@BeforeEach
		void mock() {
			request = XtaGetMessageRequest.builder()
					.messageId(MESSAGE_ID)
					.clientIdentifier(READER_IDENTIFIER)
					.build();

			messageMetaDataHolder = new Holder<>();
			responseHeaderHolder = new Holder<>();
			when(requestMapper.mapXtaGetMessageRequest(fetchRequest, authorIdentifier)).thenReturn(request);
		}

		@DisplayName("with valid request")
		@Nested
		class TestWithValidRequest {

			@BeforeEach
			void setup() {
				when(xtaMessageService.getMessage(request)).thenReturn(Optional.of(xtaMessage));

				when(responseMapper.mapMessageMetaData(xtaMessage.metaData())).thenReturn(mappedMessageMetaData);
				when(xtaMessageService.createRequestId()).thenReturn(REQUEST_ID);
				when(responseMapper.mapMsgBoxResponseTypeFromRequestId(REQUEST_ID)).thenReturn(mappedFetchResponseHeader);
				when(responseMapper.mapGenericContentContainerFromXtaMessage(xtaMessage)).thenReturn(mappedGenericContentContainer);
			}

			@DisplayName("should assign message meta data")
			@Test
			void shouldAssignMessageMetaData() {
				getMessage();

				assertThat(messageMetaDataHolder.value).isEqualTo(mappedMessageMetaData);
			}

			@DisplayName("should assign fetch response header")
			@Test
			void shouldAssignFetchResponseHeader() {
				getMessage();

				assertThat(responseHeaderHolder.value).isEqualTo(mappedFetchResponseHeader);
			}

			@DisplayName("should return")
			@Test
			void shouldReturn() {
				var genericContentContainer = getMessage();

				assertThat(genericContentContainer).isEqualTo(mappedGenericContentContainer);
			}
		}

		@DisplayName("with invalid message id")
		@Nested
		class TestWithInvalidMessageId {

			@Mock
			private InvalidMessageIDException exception;

			@DisplayName("should throw exception")
			@Test
			void shouldThrowException() {
				when(xtaMessageService.getMessage(request)).thenReturn(Optional.empty());
				when(exceptionMapper.createInvalidMessageIDException(any(), any())).thenReturn(exception);

				assertThatThrownBy(TestGetMessage.this::getMessage)
						.isEqualTo(exception);
			}
		}

		@DisplayName("with invalid request")
		@Nested
		class TestWithInvalidRequest {

			@BeforeEach
			@SneakyThrows
			void setup() {
				doThrow(new XTAWSTechnicalProblemException()).when(requestValidator).validate(request);
			}

			@DisplayName("should throw exception")
			@Test
			void shouldThrowException() {
				assertThatThrownBy(TestGetMessage.this::getMessage)
						.isInstanceOf(XTAWSTechnicalProblemException.class);
			}
		}

		@SneakyThrows
		private GenericContentContainer getMessage() {
			return msgBoxPortImpl.getMessage(fetchRequest, authorIdentifier, messageMetaDataHolder, responseHeaderHolder);
		}
	}

	@DisplayName("get status list")
	@Nested
	class TestGetStatusList {
		@Mock
		private MsgBoxStatusListRequestType statusListRequest;

		@Mock
		private PartyType authorIdentifier;

		private Holder<MsgBoxResponseType> msgBoxResponseHeader;

		@Mock
		private XtaMessageMetaDataListingRequest request;

		@BeforeEach
		void mock() {
			when(requestMapper.mapMessageMetaDataListingRequest(statusListRequest, authorIdentifier)).thenReturn(request);

			msgBoxResponseHeader = new Holder<>();
		}

		@DisplayName("with valid request")
		@Nested
		class TestWithValidRequest {
			@Mock
			private XtaMessageMetaDataListing listing;
			@Mock
			private MsgStatusListType statusListType;
			@Mock
			private MsgBoxResponseType mappedMsgBoxResponseHeader;

			@BeforeEach
			void setup() {
				when(xtaMessageService.getStatusList(request)).thenReturn(listing);
				when(responseMapper.mapMsgStatusListType(listing)).thenReturn(statusListType);
				when(responseMapper.mapMsgBoxResponseType(listing)).thenReturn(mappedMsgBoxResponseHeader);
			}

			@DisplayName("should return")
			@Test
			void shouldReturn() {
				var msgStatusList = getStatusList();

				assertThat(msgStatusList).isEqualTo(statusListType);
			}

			@DisplayName("should assign header")
			@Test
			void shouldAssignHeader() {
				getStatusList();

				assertThat(msgBoxResponseHeader.value).isEqualTo(mappedMsgBoxResponseHeader);
			}
		}

		@DisplayName("with invalid request")
		@Nested
		class TestWithInvalidRequest {

			@BeforeEach
			@SneakyThrows
			void setup() {
				doThrow(new XTAWSTechnicalProblemException()).when(requestValidator).validate(request);
			}

			@DisplayName("should throw exception")
			@Test
			void shouldThrowException() {
				assertThatThrownBy(TestGetStatusList.this::getStatusList)
						.isInstanceOf(XTAWSTechnicalProblemException.class);
			}
		}

		@SneakyThrows
		private MsgStatusListType getStatusList() {
			return msgBoxPortImpl.getStatusList(statusListRequest, authorIdentifier, msgBoxResponseHeader);
		}

	}

	@DisplayName("close")
	@Nested
	class TestClose {
		@Mock
		private MsgBoxCloseRequestType closeRequest;
		@Mock
		private PartyType authorIdentifier;

		private XtaCloseRequest request;

		@BeforeEach
		void mock() {
			request = XtaCloseRequest.builder()
					.messageId(MESSAGE_ID)
					.clientIdentifier(READER_IDENTIFIER)
					.build();
			when(requestMapper.mapXtaCloseRequest(closeRequest, authorIdentifier)).thenReturn(request);
		}

		@DisplayName("with valid request")
		@Nested
		class TestWithValidRequest {
			@BeforeEach
			void setup() {
				when(xtaMessageService.closeMessage(request)).thenReturn(true);
			}

			@DisplayName("should call close message")
			@Test
			@SneakyThrows
			void shouldCallCloseMessage() {
				close();

				verify(xtaMessageService).closeMessage(any());
			}

			@DisplayName("should return")
			@Test
			void shouldReturn() {
				close();
			}
		}

		@DisplayName("with invalid message id")
		@Nested
		class TestWithInvalidMessageId {

			@Mock
			private InvalidMessageIDException exception;

			@DisplayName("should throw exception")
			@Test
			void shouldThrowException() {
				when(xtaMessageService.closeMessage(request)).thenReturn(false);
				when(exceptionMapper.createInvalidMessageIDException(any(), any())).thenReturn(exception);

				assertThatThrownBy(TestClose.this::close)
						.isEqualTo(exception);
			}
		}

		@DisplayName("with invalid request")
		@Nested
		class TestWithInvalidRequest {
			@BeforeEach
			@SneakyThrows
			void setup() {
				doThrow(new XTAWSTechnicalProblemException()).when(requestValidator).validate(any());
			}

			@DisplayName("should throw exception")
			@Test
			void shouldThrowException() {
				assertThatThrownBy(TestClose.this::close)
						.isInstanceOf(XTAWSTechnicalProblemException.class);
			}
		}

		@SneakyThrows
		private void close() {
			msgBoxPortImpl.close(closeRequest, authorIdentifier);
		}
	}

}
